"""Callback code."""

import base64
import json
import os
import tempfile

from dash.dependencies import Input, Output, State
import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
from logzero import logger
import pandas as pd

from ..conversion import convert_seqs
from ..export import write_excel
from ..workflow import blast_and_haplotype_many, results_to_data_frames
from ..phylo import phylo_analysis
from .settings import FILE_NAME_TO_SAMPLE_NAME, SAMPLE_REGEX

from . import ui


def register_upload(app):
    @app.callback(
        Output("hidden-data", "children"),
        [Input("upload-data", "contents")],
        [State("hidden-data", "children"), State("upload-data", "filename")],
    )
    def data_uploaded(list_of_contents, hidden_data, list_of_names):
        if list_of_contents:
            with tempfile.TemporaryDirectory() as tmpdir:
                paths_reads = []
                for content, name in zip(list_of_contents, list_of_names):
                    paths_reads.append(os.path.join(tmpdir, name))
                    with open(paths_reads[-1], "wb") as tmp_file:
                        logger.info("Writing to %s", paths_reads[-1])
                        _, content = content.split(",", 1)
                        tmp_file.write(base64.b64decode(content))
                seq_files = convert_seqs(paths_reads, tmpdir, FILE_NAME_TO_SAMPLE_NAME)
                results = blast_and_haplotype_many(paths_reads)
                df_summary, df_blast, df_haplotyping = results_to_data_frames(results, SAMPLE_REGEX)

                row_select = (df_summary.orig_sequence != "-") & (df_summary.region != "-")
                columns = ["query", "region", "orig_sequence"]
                phylo_result = phylo_analysis(df_summary[row_select][columns])
            return json.dumps(
                {
                    "summary": df_summary.to_dict(),
                    "blast": df_blast.to_dict(),
                    "haplotyping": df_haplotyping.to_dict(),
                    "phylo": phylo_result,
                }
            )


def load_hidden_data(hidden_data):
    raw_data = json.loads(hidden_data)

    def decode(data, key):
        if key in ("summary", "blast", "haplotyping"):
            return pd.DataFrame.from_dict(data)
        else:
            return data

    return {key: decode(raw_data[key], key) for key in raw_data}


def register_computation_complete(app):
    @app.callback(Output("page-content", "children"), [Input("hidden-data", "children")])
    def computation_complete(hidden_data):
        if not hidden_data:
            return ui.render_page_content_empty_children()
        else:
            data = load_hidden_data(hidden_data)
            mime = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
            mime = "application/octet-stream"
            with tempfile.NamedTemporaryFile() as tmpf:
                # df_summary, df_blast, df_haplotyping, path):
                write_excel(data["summary"], data["blast"], data["haplotyping"], tmpf.name)
                tmpf.seek(0)
                xlsx = base64.b64encode(tmpf.read()).decode("utf-8")
            return [
                html.P(
                    children=[
                        html.A(
                            children=[
                                html.I(className="fas fa-file-excel ml-2 mr-2"),
                                "Download XLSX",
                            ],
                            # href="data:text/html,<script>alert('hi');</script>",
                            download="hlso_result.xlsx",
                            target="_blank",
                            href="data:%s;base64,%s" % (mime, xlsx),
                        )
                    ]
                ),
                dbc.Tabs(
                    children=[
                        dbc.Tab(ui.render_tab_summary(data), label="Summary", id="tab-summary"),
                        dbc.Tab(ui.render_tab_blast(data), label="BLAST", id="tab-blast"),
                        dbc.Tab(
                            ui.render_tab_haplotyping(data),
                            label="Haplotyping",
                            id="tab-haplotyping",
                        ),
                        dbc.Tab(
                            ui.render_tab_dendrograms(data),
                            label="Dendrograms",
                            id="tab-dendgrograms",
                        ),
                    ]
                ),
            ]


def register_row_clicks(app):
    @app.callback(
        Output("blast-current-match", "children"),
        [Input("hidden-data", "children"), Input("blast-table", "selected_row_ids")],
    )
    def update_haplotype_match(hidden_data, selected_row_ids):
        # logger.info("Selected %s from %s", selected_row_ids, hidden_data)
        if selected_row_ids and hidden_data:
            selected_row_ids = list(map(str, selected_row_ids))[0]
            data = load_hidden_data(hidden_data)
            df_blast = data["blast"]
            alignment = df_blast.loc[selected_row_ids].alignment
            ncbi_tpl = "%s?DATABASE=nt&PROGRAM=blastn&MEGABLAST=on&QUERY=>%s%%0A%s"
            ncbi_url = ncbi_tpl % (
                "https://blast.ncbi.nlm.nih.gov/Blast.cgi",
                df_blast.loc[selected_row_ids]["query"],
                df_blast.loc[selected_row_ids].orig_sequence,
            )
            return [
                html.P(
                    children=[
                        html.Div(
                            children=[
                                html.A(
                                    children=[
                                        html.I(className="fas fa-external-link-alt"),
                                        " RunNCBI BLAST for this sequence",
                                    ],
                                    href=ncbi_url,
                                )
                            ],
                            className="mt-3",
                        )
                    ]
                ),
                dcc.Markdown("```text\n%s\n```" % alignment, className="mt-3"),
            ]
        return [html.P(["no match selected yet"])]



"""Download of seed references."""

import os
import shutil
import tempfile

from logzero import logger

from .cli import _proc_args
from .conversion import convert_seqs


def run(parser, args):
    """Perform conversion."""
    args = _proc_args(parser, args)
    logger.info("Converting sequences...")
    logger.info("Args = %s", args)
    with tempfile.TemporaryDirectory() as tmpdir:
        seq_files = convert_seqs(args.seq_files, tmpdir)
        for seq_file in seq_files:
            shutil.copy(seq_file, os.path.join(args.out_dir, os.path.basename(seq_file)))
    logger.info("All done, have a nice day!")


def add_parser(subparser):
    """Configure the ``argparse`` sub parser."""
    parser = subparser.add_parser("convert")
    parser.set_defaults(func=run)

    parser.add_argument(
        "--file-name-as-seq-name",
        action="store_true",
        default=False,
        help="Set file name to sample name",
    )
    parser.add_argument("out_dir", help="Path to output directory.")
    parser.add_argument("seq_files", nargs="+", default=[], action="append")



"""Perform consensus building for creating the reference step.

The following actions are performed for this step:

1. Build consensus sequence for each seed:  for each seed, consider the NCBI WWWBLAST result
   and build replace the seed segment by the database hit and perform an end-to-end MSA for
   this for each seed sequence.Generate consensus seed sequence for this.

2. Perform a BLAST search of the consensus seed sequences and generate type reference sequences
   for each haplotype.

3. Perform MSA of the type reference sequences and generate a haplotyping table from this.
"""

# TODO: replace dict representing variant by attrs-based class.

import itertools
import os
import shlex
import shutil
import subprocess
import tempfile
import textwrap

from Bio.Blast import NCBIXML
from Bio.Align import Applications
from logzero import logger

from .ref_download import REF_SEQS
from .ref_blast import load_tsv
from .cli import _proc_args
from .common import call_variants, describe, load_fasta, normalize_var, only_bases
from .paste import REF_FILE, do_paste
from .workflow import only_blast


def paste_query_seq(query_seq, blast_alignment):
    """Paste database sequence from ``blast_alignment`` into ``query_seq``."""
    hsp = blast_alignment.hsps[0]

    if hsp.strand[0] == "Plus":  # query strand is forward
        prefix = query_seq[: hsp.query_start]
        sbjct_seq = "".join([c for c in hsp.sbjct if c in "cgatnCGATN"])
        suffix = query_seq[hsp.query_end :]
        seq = prefix + sbjct_seq + suffix
    else:
        raise Exception("Cannot handle matches on query reverse strand")

    return "".join([c for c in seq if c != "N"]).upper()


def consensus(cs):
    counter = {}
    for c in cs:
        counter.setdefault(c, 0)
        counter[c] += 1
    max_val = max(counter.values())
    for key, value in counter.items():
        if value == max_val:
            return key
    return "N"


def write_consensus(path, query_name, query_seq, alignments, args):
    if os.path.exists(path):
        logger.warn("Consensus sequence exists, skipping: %s", path)
        return

    logger.info("Creating consensus for %s", query_name)
    single_hsps = [a for a in alignments if len(a.hsps) == 1]
    logger.info("- removed %d alignments with HSP count != 1", len(alignments) - len(single_hsps))
    sorted_alis = sorted(
        single_hsps, key=lambda a: a.hsps[0].identities / a.hsps[0].align_length, reverse=True
    )

    good_alis = [
        a for a in sorted_alis if a.hsps[0].identities >= a.hsps[0].align_length - args.max_errors
    ]
    logger.info("- picked %d alignments with up to %d errors", len(good_alis), args.max_errors)
    logger.info("  %s", [a.title.split("|")[3] for a in good_alis])
    if not good_alis:
        logger.error("Could not pick any alignments for %s. Skipping", query_name)
        return

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_fasta = os.path.join(tmpdir, "seqs.fasta")
        logger.info("Writing to %s...", tmp_fasta)
        with open(tmp_fasta, "wt") as tmpf:
            for ali_count, ali in enumerate(good_alis):
                pasted_seq = "\n".join(textwrap.wrap(paste_query_seq(query_seq, ali)))
                print(">%s\n%s" % (ali.title.split("|")[3], pasted_seq), file=tmpf)

        path_alignment = path[: -len(".consensus.fasta")] + ".clustalw.fasta"
        if os.path.exists(path_alignment):
            logger.warn("Already exists: %s", path_alignment)
            logger.warn(" => not running ClustalW")
        elif ali_count == 0:
            logger.warn("Only one sequence, no need to align, just copy")
            shutil.copy(tmp_fasta, path_alignment)
        else:
            logger.info("Running ClustalW")
            cmd = [
                "clustalw",
                "-INFILE=%s" % tmp_fasta,
                "-ALIGN",
                "-OUTFILE=%s" % path_alignment,
                "-OUTPUT=FASTA",
                "-OUTORDER=INPUT",
                "-SEED=42",
            ]
            logger.info("- %s", " ".join(cmd))
            res = subprocess.check_call(cmd)
            if res != 0:
                logger.error("Runing ClustalW failed with retval %d", res)
            else:
                logger.info("Done running ClustalW!")

    logger.info("Computing consensus sequence to %s", path)
    fasta = load_fasta(path_alignment)
    consensus_seq = "".join(filter(lambda x: x in "ACGTN", map(consensus, zip(*fasta.values()))))
    with open(path, "wt") as outputf:
        print(
            ">consensus-%s\n%s" % (query_name, "\n".join(textwrap.wrap(consensus_seq))),
            file=outputf,
        )


def build_seed_consensus(records, args):
    """Build consensus sequence for each seed."""
    logger.info("Building seed consensus sequences...")

    base_dir = os.path.dirname(args.in_tsv)

    for record in records:
        logger.info("Handling %s", record["path"])
        basename = os.path.basename(record["path"])[: -len(".fasta")]
        fasta = load_fasta(os.path.join(base_dir, basename + ".fasta"))

        path_blast = os.path.join(base_dir, basename + ".blast.xml")
        with open(path_blast, "rt") as inputf:
            for record in NCBIXML.parse(inputf, debug=1 if args.verbose else 0):
                # Compute consensus for each query sequence if it does not exist already.
                path_consensus = os.path.join(base_dir, basename + ".consensus.fasta")
                if os.path.exists(path_consensus):
                    logger.warn("Skipping, already exists: %s", path_consensus)
                    continue
                query_name = record.query.split()[0]
                write_consensus(
                    path=path_consensus,
                    query_name=query_name,
                    query_seq=fasta[query_name],
                    alignments=record.alignments,
                    args=args,
                )


def build_haplotype_sequences(records, args):
    """Build consensus sequence for each seed."""
    logger.info("Building building haplotype sequences...")

    ref_seqs = load_fasta(REF_FILE)
    base_dir = os.path.dirname(args.in_tsv)

    results = {}  # actually consensus seeds
    for record in records:
        out_path = os.path.join(base_dir, record["region"] + ".fasta")
        if os.path.exists(out_path):
            logger.warn("Output path exists, skipping: %s", out_path)
            continue

        key = (record["haplotype"], record["region"])
        logger.info("Handling %s", record["path"])
        basename = os.path.basename(record["path"])[: -len(".fasta")]
        matches = only_blast(os.path.join(base_dir, basename + ".consensus.fasta"))
        for match in matches:
            if match.database is None:
                logger.info("  => no match for %s", match.query)
            else:
                logger.info("Pasting region=%s haplotype=%s", record["region"], record["haplotype"])
                seq = do_paste(match, ref_seqs)
                results.setdefault(record["region"], {})[record["haplotype"]] = seq

    for region, region_results in results.items():
        fasta = os.path.join(base_dir, region + ".fasta")
        logger.info("Writing padded consensus seed to to %s", fasta)
        with open(fasta, "wt") as outputf:
            for haplotype, seq in region_results.items():
                wrapped = textwrap.wrap(seq)
                print(">%s_%s\n%s" % (region, haplotype, "\n".join(wrapped)), file=outputf)

    return results


def strip_n(s):
    return "".join([c for c in s if c not in "nN"])


def build_haplotyping_table(records, args):
    """Build consensus sequence for each seed."""
    logger.info("Building haplotyping table...")

    ref_seqs = load_fasta(REF_FILE)
    base_dir = os.path.dirname(args.in_tsv)

    ali_input = {
        key.split("_")[1]: {key.split("_")[1]: strip_n(value)} for key, value in ref_seqs.items()
    }

    for region, seqs in ali_input.items():
        fasta_path = os.path.join(base_dir, "%s.fasta" % region)
        if not os.path.exists(fasta_path):
            logger.warn("Consensus sequences FASTA not found, skipping: %s", fasta_path)
        else:
            logger.info("Loading consensus sequences FASTA: %s", fasta_path)
            seqs.update(load_fasta(fasta_path))

    result = {}

    for region, seqs in ali_input.items():
        for seq_name, ref_seq in ref_seqs.items():
            ref_name, ref_region = seq_name.split("_", 1)
            if ref_region == region:
                ref_offset = 0
                for ref_offset, c in enumerate(ref_seq):
                    if c == "N":
                        break
                break  # => ref_name, ref_seq, ref_offset are set, else Exception is raised
        else:
            raise Exception("Could not find reference for %s" % ref_name)

        path_alignment = os.path.join(base_dir, "%s.clustalw.fasta" % region)
        if len(seqs) == 1:
            logger.warn("Only reference and no consensus, skipping: %s", path_alignment)
            continue

        logger.info("Running ClustalW on consensus alignments for %s", region)
        if os.path.exists(path_alignment):
            logger.warn("Ref+consensus alignment already exists, skipping: %s", path_alignment)
        else:
            with tempfile.TemporaryDirectory() as tmpdir:
                tmp_fasta = os.path.join(tmpdir, "seqs.fasta")
                logger.info("Writing to %s...", tmp_fasta)
                with open(tmp_fasta, "wt") as tmpf:
                    for name, seq in seqs.items():
                        print(">%s\n%s" % (name, seq), file=tmpf)

                logger.info("Running ClustalW")
                cmd = [
                    "clustalw",
                    "-INFILE=%s" % tmp_fasta,
                    "-ALIGN",
                    "-OUTFILE=%s" % path_alignment,
                    "-OUTPUT=FASTA",
                    "-OUTORDER=INPUT",
                    "-SEED=42",
                ]
                logger.info("- %s", " ".join(cmd))
                res = subprocess.check_call(cmd)
                if res != 0:
                    logger.error("Runing ClustalW failed with retval %d", res)
                else:
                    logger.info("Done running ClustalW!")

        logger.info("Loading MSA from %s", path_alignment)
        alignment = load_fasta(path_alignment)

        logger.info("Generating variant table for each haplotype")
        variants = {
            key: call_variants(alignment[region], alignment[key], REF_SEQS[region]["start"])
            for key in sorted(alignment.keys())
            if key != region
        }
        xs = set()
        for calls in variants.values():
            for vals in calls.values():
                pos = vals["pos"]
                ref = only_bases(vals["ref_bases"])
                alt = only_bases(vals["alt_bases"])
                desc = vals["description"]
                xs.add((pos, ref, alt, desc))

        table = {}
        haplotypes = [h.split("_")[1] for h in variants.keys()]

        header = ["row", "pos", "ref_bases", "alt_bases", "description"] + haplotypes
        lines = []
        for row, (pos, ref, alt, desc) in enumerate(sorted(xs), 1):
            ref_seq, region = seq_name.split("_", 1)
            record = {"reference": ref_seq, "region": region, "position": pos}
            line = [row, pos, ref, alt, desc]

            # Find reference allele bases.
            ref = "N"
            for name in variants.keys():
                if pos in variants[name] and variants[name][pos]["description"] == desc:
                    ref = only_bases(variants[name][pos]["ref"])
                    break

            for name in variants.keys():
                if pos in variants[name] and variants[name][pos]["description"] == desc:
                    record[name.split("_", 1)[1]] = only_bases(variants[name][pos]["alt"])
                    line.append("+")
                else:
                    record[name.split("_", 1)[1]] = only_bases(ref)
                    line.append("-")
            table[(pos, ref, alt, desc)] = record
            lines.append(line)

        result[(ref_name, ref_region)] = dict(sorted(table.items()))

        print("Haplotyping Table")
        print("-----------------")
        print()
        print("\t".join(map(str, header)).replace("_bases", "").upper() + "\n")
        for line in lines:
            print("\t".join(map(str, line)))

    if args.output_table:
        logger.info("Writing haplotyping table to %s", args.output_table)
        separator = (
            "# --------\t-------\t------\t------\t------\t---------------"
            + "\t------" * len(haplotypes)
        )
        with open(args.output_table, "wt") as haplof:
            print(separator, file=haplof)
            print(
                "\t".join(["reference", "region", "pos", "ref", "alt", "description"] + haplotypes),
                file=haplof,
            )
            print(separator, file=haplof)
            for (ref_name, ref_region), records in result.items():
                for (pos, ref, alt, desc), record in records.items():
                    line = [ref_name, ref_region, pos, ref, alt, desc] + [
                        record[h] for h in haplotypes
                    ]
                    print("\t".join(map(str, line)), file=haplof)
                print(separator, file=haplof)
    else:
        logger.info("Writing no haplotype table (use --output-table to specify the path")


def run(parser, args):
    """Perform the consensus computation."""
    logger.info("Finishing reference computation.")
    logger.info("Arguments are %s", vars(args))

    _header, records = load_tsv(args.in_tsv)

    build_seed_consensus(records, args)
    build_haplotype_sequences(records, args)
    build_haplotyping_table(records, args)

    logger.info("All done. Have a nice day!")


def add_parser(subparser):
    """Configure the ``argparse`` sub parser."""
    parser = subparser.add_parser("ref_consensus")
    parser.set_defaults(func=run)

    parser.add_argument("--verbose", action="store_true", default=False, help="Enable verbose mode")
    parser.add_argument(
        "--max-errors",
        type=int,
        default=0,  # TODO: make sequence-specific!
        help="Maximal number of mismatches to accept in seed consensus computation",
    )
    parser.add_argument("--output-table", default=None, help="Path to output haplotype table.")
    parser.add_argument("in_tsv", help="Path to output TSV file.")



"""Classification of C. liberibacter solanacearum following IPPC standards.

This module is the main entry point.
"""

import argparse
from subprocess import check_output, CalledProcessError
import sys

from . import cli, web, paste, convert, ref_download, ref_blast, ref_consensus

#: The required executables in PATH.
REQUIRED_EXECTUABLES = ("blastn", "makeblastdb")


def main(argv=None):
    """Main entrypoint (before parsing command line arguments)."""
    parser = argparse.ArgumentParser(description="Classify Lso Sanger reads.")
    subparsers = parser.add_subparsers()
    convert.add_parser(subparsers)
    cli.add_parser(subparsers)
    web.add_parser(subparsers)
    paste.add_parser(subparsers)
    ref_download.add_parser(subparsers)
    ref_blast.add_parser(subparsers)
    ref_consensus.add_parser(subparsers)

    for prog in REQUIRED_EXECTUABLES:
        found_all = True
        try:
            check_output(["which", prog])
        except CalledProcessError:
            found_all = False
            print("ERROR: Required program %s not found!" % repr(prog), file=sys.stderr)
    if not found_all:
        parser.exit(1)

    args = parser.parse_args(argv)
    if not hasattr(args, "func"):
        parser.print_help()
        parser.exit(1)
    else:
        return args.func(parser, args)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""Code for exporting the data frames generated from the ``workflow`` module."""

import pandas as pd

from .web.settings import (
    MIN_IDENTITY_GREEN,
    MIN_IDENTITY_YELLOW,
    FONT_COLOR_GREEN,
    FONT_COLOR_YELLOW,
    FONT_COLOR_RED,
    BG_COLOR_GREEN,
    BG_COLOR_YELLOW,
    BG_COLOR_RED,
)

#: Sheet name for summary sheet.
SHEET_SUMMARY = "Summary"
#: Sheet name for BLAST results.
SHEET_BLAST = "BLAST"
#: Sheet name for haplotyping results.
SHEET_HAPLOTYPING = "Haplotyping"


def write_excel(df_summary, df_blast, df_haplotyping, path):
    """Write the sheets to the given ``path``."""
    writer = pd.ExcelWriter(path, engine="xlsxwriter")

    # Write out data frames.
    df_summary.to_excel(writer, sheet_name=SHEET_SUMMARY, index=False)
    df_blast.to_excel(writer, sheet_name=SHEET_BLAST, index=False)
    df_haplotyping.to_excel(writer, sheet_name=SHEET_HAPLOTYPING, index=False)

    # Get workbook and sheets.
    workbook = writer.book
    sheet_summary = writer.sheets[SHEET_SUMMARY]
    sheet_blast = writer.sheets[SHEET_BLAST]

    # Setup formats.
    format_green = workbook.add_format({"bg_color": BG_COLOR_GREEN, "font_color": FONT_COLOR_GREEN})
    format_yellow = workbook.add_format(
        {"bg_color": BG_COLOR_YELLOW, "font_color": FONT_COLOR_YELLOW}
    )
    format_red = workbook.add_format({"bg_color": BG_COLOR_RED, "font_color": FONT_COLOR_RED})
    for f in format_green, format_yellow, format_red:
        f.set_num_format("0.0")

    # Setup conditional formats.
    cond_green = {
        "type": "cell",
        "criteria": "between",
        "maximum": 100,
        "minimum": MIN_IDENTITY_GREEN,
        "format": format_green,
    }
    cond_yellow = {
        "type": "cell",
        "criteria": "between",
        "maximum": MIN_IDENTITY_GREEN,
        "minimum": MIN_IDENTITY_YELLOW,
        "format": format_yellow,
    }
    cond_red = {
        "type": "cell",
        "criteria": "between",
        "maximum": MIN_IDENTITY_YELLOW,
        "minimum": 0,
        "format": format_red,
    }

    # Apply conditional formatting.
    c_summary = chr(ord("A") + df_summary.columns.get_loc("identity"))
    c_blast = chr(ord("A") + df_blast.columns.get_loc("identity"))
    for cond in cond_green, cond_yellow, cond_red:
        sheet_summary.conditional_format(
            "%s2:%s%d" % (c_summary, c_summary, df_summary.shape[0] + 1), cond
        )
        sheet_blast.conditional_format("%s2:%s%d" % (c_blast, c_blast, df_blast.shape[0] + 1), cond)

    writer.save()



"""Code for reading the haplotyping table information.

This contains the informative positions for haplotyping of calls.
"""

import os
import shlex
import subprocess
import tempfile
import typing

import attr
import json
from logzero import logger

from .blast import BlastMatch
from .common import call_variants, normalize_var


@attr.s(auto_attribs=True, frozen=True)
class HaplotypingPos:
    """A haplotyping position."""

    #: database reference sequence name
    reference: str
    #: 0-based position on database reference
    position: int
    #: mapping from haplotype name to value
    haplo_values: typing.Dict[str, str]


def load_haplotyping_table(path: str) -> typing.Dict[typing.Tuple[str, int], HaplotypingPos]:
    """Load haplotyping table from the given ``path``."""
    # logger.debug("Loading haplotyping table from %s", path)
    result = {}

    if not os.path.exists(path):
        logger.warning("Could not load haplotyping table from %s", path)
        return result

    header = None
    with open(path, "rt") as inputf:
        for line in inputf:
            if line.startswith("#"):
                continue
            arr = line.rstrip().split("\t")
            if not header:
                header = arr
            else:
                record = dict(zip(header, arr))
                key = (record["reference"], int(record["pos"]) - 1)
                result[key] = HaplotypingPos(
                    reference=key[0], position=key[1], haplo_values=dict(zip(header[2:], arr[2:]))
                )
    # logger.debug("Done loading %d records", len(result))
    return result


#: The haplotype table.
HAPLOTYPE_TABLE = load_haplotyping_table(
    os.path.join(os.path.dirname(__file__), "data", "haplotype_table.txt")
)
# logger.debug("haplotype table = %s", HAPLOTYPE_TABLE)

#: The haplotype names
HAPLOTYPE_NAMES = "ABCDE"


@attr.s(auto_attribs=True, frozen=True)
class HaplotypingResult:
    """A haplotyping result."""

    #: The file name used for haplotyping
    filename: str
    #: The query name
    query: str
    #: mapping from ``(reference, zero_based_pos)`` to allele value
    informative_values: typing.Dict[typing.Tuple[str, int], str]

    def merge(
        self, other: typing.TypeVar("HaplotypingResult")
    ) -> typing.TypeVar("HaplotypingResult"):
        keys = list(
            sorted(set(self.informative_values.keys()) | set(other.informative_values.keys()))
        )
        merged = {}
        for key in keys:
            if key in self.informative_values and key in other.informative_values:
                here = self.informative_values[key]
                there = other.informative_values[key]
                if here == there:
                    merged[key] = here
            merged[key] = self.informative_values.get(key, other.informative_values.get(key))
        return HaplotypingResult(filename="-", query="-", informative_values=merged)  # post-merging

    def asdict(self, only_summary=False) -> typing.Dict:
        informative = {}
        scores = {}
        for name in HAPLOTYPE_NAMES:
            plus, minus = self.compare(name)
            informative["%s_pos" % name] = plus
            informative["%s_neg" % name] = minus
            scores[name] = plus - minus
        best_score = max(scores.values())
        if best_score > 0:
            best_haplotypes = ",".join(
                [key for key, value in scores.items() if value == best_score]
            )
        else:
            best_haplotypes = "-"
        if only_summary:
            return {"best_haplotypes": best_haplotypes, "best_score": best_score}
        else:
            return {
                "filename": self.filename,
                "best_haplotypes": best_haplotypes,
                "best_score": best_score,
                **informative,
                **{
                    "%s:%d" % (key[0], key[1] + 1): self.informative_values.get(key)
                    for key in HAPLOTYPE_TABLE
                },
            }

    def compare(self, haplotype: str) -> typing.Tuple[int, int]:
        """Return ``(match_count, mismatch_count)`` for the given ``haplotype``."""
        positive = 0
        negative = 0
        for key, value in self.informative_values.items():
            if HAPLOTYPE_TABLE[key].haplo_values[haplotype] == value:
                positive += 1
            else:
                negative += 1
        return (positive, negative)

    @classmethod
    def fromdict(self, dict_: typing.Dict) -> typing.TypeVar("HaplotypingResult"):
        informative_values = {}
        for key, value in dict_.items():
            if ":" in key and value is not None:
                arr = key.split(":", 1)
                informative_values[(arr[0], int(arr[1]) - 1)] = value
        return HaplotypingResult(
            filename=dict_["filename"], query=dict_["query"], informative_values=informative_values
        )


@attr.s(auto_attribs=True, frozen=True)
class HaplotypingResultWithMatches:
    """Result from haplotyping."""

    #: The actual ``HaplotypingResult``.
    result: typing.Optional[HaplotypingResult]
    #: The ``BlastMatch``es that the haplotyping is based on.
    matches: typing.Optional[typing.Tuple[BlastMatch]]

    @staticmethod
    def build_empty() -> typing.TypeVar("HaplotypingResultWithMatches"):
        return HaplotypingResultWithMatches(None, None)


def run_haplotyping(
    matches: typing.Iterable[BlastMatch]
) -> typing.Dict[str, HaplotypingResultWithMatches]:
    """Perform the haplotyping based on the match."""
    results_matches = {}
    results_haplo = {}

    # TODO: properly handle overlapping changes; directly look into alignment
    for match in matches:
        ref = match.database
        if "_" in ref:
            ref = ref.split("_")[0]

        calls = call_variants(match.alignment.hseq, match.alignment.qseq, match.database_start)

        informative_values = {}
        for (h_ref, h_pos), variant in HAPLOTYPE_TABLE.items():
            if ref == h_ref and h_pos >= match.database_start and h_pos < match.database_end:
                if h_pos + 1 in calls:
                    informative_values[(ref, h_pos)] = variant.haplo_values["alt"]
                else:
                    informative_values[(ref, h_pos)] = variant.haplo_values["ref"]

        result = HaplotypingResult(
            filename=match.path, query=match.query, informative_values=informative_values
        )
        if result.filename in results_haplo:
            results_matches[result.filename].append(match)
            results_haplo[result.filename] = results_haplo[result.filename].merge(result)
        else:
            results_matches[result.filename] = [match]
            results_haplo[result.filename] = result

    return {
        filename: HaplotypingResultWithMatches(
            result=results_haplo[filename], matches=results_matches[filename]
        )
        for filename in results_matches
    }



"""Setup of Haplotype-Lso Dash application."""

import os

import dash
import flask

from . import callbacks, settings
from .ui import build_layout
from hlso import __version__

#: Path to assets.
ASSETS_FOLDER = os.path.join(os.path.dirname(__file__), "assets")

#: The Flask application to use.
app_flask = flask.Flask(__name__)

# Setup URL prefix for Flask.
app_flask.config["APPLICATION_ROOT"] = "%s/" % settings.PUBLIC_URL_PREFIX

#: The Dash application to run.
app = dash.Dash(
    __name__,
    # Use our specific Flask app
    server=app_flask,
    # All files from "assets/" will be served as "/assets/*"
    assets_folder=ASSETS_FOLDER,
    # The visualization will be served below "/dash"
    routes_pathname_prefix="/dash/",
    requests_pathname_prefix="%s/dash/" % settings.PUBLIC_URL_PREFIX,
)

# Set app title
app.title = "hlso v%s" % __version__

# Initialize UI
app.layout = build_layout()

# Serve assets locally
app.css.config.serve_locally = True
app.scripts.config.serve_locally = True

# TODO: register callbacks
callbacks.register_upload(app)
callbacks.register_computation_complete(app)
callbacks.register_row_clicks(app)

# Add redirection for root.
@app_flask.route("/")
def redirect_root():
    return flask.redirect("%s/dash/" % settings.PUBLIC_URL_PREFIX)



"""Code for the command line interface to ``hlso``."""

import tempfile
import typing

import attr
from logzero import logger

from .conversion import convert_seqs
from .export import write_excel
from .phylo import phylo_analysis
from .workflow import blast_and_haplotype_many, results_to_data_frames

from .web.settings import SAMPLE_REGEX


def _proc_args(parser, args):
    seq_files = []
    for lst in args.seq_files:
        seq_files += lst
    args.seq_files = seq_files
    return args


@attr.s(auto_attribs=True, frozen=True)
class Config:
    """Configuration of the command line interface."""

    #: The paths to the input files.
    input_paths: typing.Tuple[str]
    #: The path to the output file.
    output_path: str
    #: Whether or not the sample names are to be inferred from the command line.
    #: If not, then the sequence names are used.
    sample_name_from_file: bool = False
    #: The regular expression to parse information from the sample.
    sample_regex: str = SAMPLE_REGEX


def run(parser, args):
    """Run the ``hlso`` command line interface."""
    args = _proc_args(parser, args)
    config = Config(
        input_paths=tuple(args.seq_files),
        output_path=args.output,
        sample_name_from_file=args.sample_name_from_file,
        sample_regex=args.sample_regex,
    )
    logger.info("Starting Lso classification.")
    logger.info("Arguments are %s", config)
    with tempfile.TemporaryDirectory() as tmpdir:
        logger.info("Converting sequences (if necessary)...")
        seq_files = convert_seqs(args.seq_files, tmpdir, config.sample_name_from_file)
        config = Config(**{**attr.asdict(config), "input_paths": tuple(sorted(seq_files))})
        logger.info("Running BLAST and haplotyping...")
        results = blast_and_haplotype_many(seq_files)
        logger.info("Converting results into data frames...")
        df_summary, df_blast, df_haplotyping = results_to_data_frames(results, args.sample_regex)
        logger.info("Summary:\n%s", df_summary)
        logger.info("Writing XLSX file to %s", args.output)
        write_excel(df_summary, df_blast, df_haplotyping, args.output)
        if "region" in df_summary.columns:
            row_select = (df_summary.orig_sequence != "-") & (df_summary.region != "-")
            columns = ["query", "region", "orig_sequence"]
            dendro_out = args.output[: -len(".xlsx")] + ".%s.png"
            phylo_analysis(df_summary[row_select][columns], path_out=dendro_out)
        else:
            logger.info("Column 'region' not in summary, not computing similarities")
    logger.info("All done. Have a nice day!")


def add_parser(subparser):
    """Configure the ``argparse`` sub parser."""
    parser = subparser.add_parser("cli")
    parser.set_defaults(func=run)

    parser.add_argument(
        "--sample-name-from-file",
        action="store_true",
        default=False,
        help="Use sample name instead of file name",
    )
    parser.add_argument(
        "--sample-regex",
        default=SAMPLE_REGEX,
        help="Regular expression to match file name to sample name.",
    )
    parser.add_argument("-o", "--output", default="clsified.xlsx", help="Path to output file")
    parser.add_argument("seq_files", nargs="+", default=[], action="append")



"""Prepare modified copies of reference sequence by inserting BLAST matches into them."""

import os
import tempfile
import textwrap

from logzero import logger

from .common import load_fasta, revcomp
from .conversion import convert_seqs
from .workflow import blast_and_haplotype_many, REF_FILE
from .cli import _proc_args


def do_paste(match, ref_seqs=None):
    ref_seqs = ref_seqs or load_fasta(REF_FILE)
    seq = ref_seqs[match.database]
    qseq = "".join([c for c in match.alignment.qseq if c in "cgatnCGATN"])
    if match.query_strand == "-":
        qseq = revcomp(seq)
    seq = seq[: match.database_start] + qseq + seq[match.database_end :]
    return "".join([c for c in seq if c != "N"]).upper()


def write_pasted(results, output_prefix):
    logger.info("Loading reference sequences...")
    ref_seqs = load_fasta(REF_FILE)
    logger.info("Writing pasted sequences...")
    for result in results:
        for match in result["matches"]:
            if match.database is None:
                logger.info("  => no match for %s", match.query)
                continue
            query = match.query.split(" ")[0]
            out_path = "{}{}/{}-{}.fasta".format(
                output_prefix, match.database, query, match.database
            )
            if not os.path.exists(os.path.dirname(out_path)):
                os.makedirs(os.path.dirname(out_path))
            logger.info("  - %s", out_path)
            with open(out_path, "wt") as outputf:
                seq = do_paste(match, ref_seqs)
                # print(">{}-{}".format(query, match.database), file=outputf)
                print(">{} ({} bp)".format(query, len(seq)), file=outputf)
                print("\n".join(textwrap.wrap(seq)), file=outputf)


def run(parser, args):
    """Run the ``hlso`` command line interface."""
    args = _proc_args(parser, args)
    logger.info("Starting Lso classification.")
    logger.info("Arguments are %s", vars(args))
    with tempfile.TemporaryDirectory() as tmpdir:
        logger.info("Converting sequences (if necessary)...")
        seq_files = convert_seqs(args.seq_files, tmpdir)
        logger.info("Running BLAST (and haplotyping)...")
        results = blast_and_haplotype_many(seq_files)
        logger.info("Writing out pasted sequence")
        write_pasted(results, args.output_prefix)
    logger.info("All done. Have a nice day!")


def add_parser(subparser):
    """Configure the ``argparse`` sub parser."""
    parser = subparser.add_parser("paste")
    parser.set_defaults(func=run)

    parser.add_argument(
        "--keep-masked",
        default=False,
        help="Keep masked reference sequence (mostly useful for debugging purposes).",
    )
    parser.add_argument(
        "-o", "--output-prefix", default="hlso_paste_out.d/", help="Prefix for output files"
    )
    parser.add_argument("seq_files", nargs="+", default=[], action="append")



"""Running seeds through WWWBLAST."""

from multiprocessing import Pool
import os
import textwrap

from Bio.Blast import NCBIWWW
from logzero import logger

from .ref_download import load_tsv
from .cli import _proc_args


def do_blast(x):
    record, args = x

    out_dir = os.path.dirname(args.in_tsv)
    out_file = os.path.basename(record["path"])[: -len(".fasta")] + ".blast.xml"
    out_path = os.path.join(out_dir, out_file)

    logger.info("Loading %s", record["path"])
    with open(os.path.join(out_dir, record["path"]), "rt") as inputf:
        fasta = inputf.read()

    logger.info("Running NCBI WWW BLAST for %s", fasta.split()[0][1:])
    blast_res = NCBIWWW.qblast(program="blastn", database="nt", sequence=fasta, megablast=True)
    with open(out_path, "wt") as outputf:
        logger.info("Writing NCBI WWW BLAST result for %s", fasta.split()[0][1:])
        outputf.write(blast_res.read())


def run(parser, args):
    """Perform NCBIWWW of the seeds."""
    logger.info("Starting NCBI WWW BLAST of seed sequences and reference.")
    logger.info("Arguments are %s", vars(args))

    out_dir = os.path.dirname(args.in_tsv)

    jobs = []

    logger.info("Running NCBI WWW BLAST for each sequence...")
    header, records = load_tsv(args.in_tsv)
    for record in records:
        out_file = os.path.basename(record["path"])[: -len(".fasta")] + ".blast.xml"
        out_path = os.path.join(out_dir, out_file)
        if os.path.exists(out_path):
            logger.warn("Output path already exists: %s", out_path)
            continue
        else:
            jobs.append((record, args))

    with Pool(args.num_threads) as p:
        p.map(do_blast, jobs, chunksize=1)

    logger.info("All done. Have a nice day!")


def add_parser(subparser):
    """Configure the ``argparse`` sub parser."""
    parser = subparser.add_parser("ref_blast")
    parser.set_defaults(func=run)

    parser.add_argument(
        "--num-threads", default=8, type=int, help="Number of parallel searches to run"
    )
    parser.add_argument("in_tsv", help="Path to input TSV file.")



"""Phylogenetics analysis"""

import os
import tempfile
import typing

from logzero import logger
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from scipy.spatial import distance
from scipy.cluster import hierarchy

from .blast import run_blast, run_makeblastdb
from .common import write_fasta, load_fasta


def phylo_analysis(
    df: pd.DataFrame, *, path_out: typing.Optional[str] = None
) -> typing.Dict[str, typing.Dict[str, object]]:
    logger.info("Performing phylogenetics analysis on\n%s", df)
    result = {}

    for region, group in df.groupby("region"):
        seqs = dict(zip(group["query"], group["orig_sequence"]))
        path_ref = os.path.join(os.path.dirname(__file__), "reference/%s.fasta" % region)
        if os.path.exists(path_ref):
            logger.info("Loading reference %s", path_ref)
            seqs.update(load_fasta(path_ref))
        keys = tuple(sorted(seqs.keys()))

        if len(seqs) == 1:  # skip if only one sequence given
            result[region] = {"message": "Cannot compute dendrogram for one sequence."}
            continue

        with tempfile.TemporaryDirectory() as tmp_dir:
            logger.info("Performing phylogenetics analysis for %s region", region)
            path_seqs = os.path.join(tmp_dir, "seqs.fasta")
            with open(path_seqs, "wt") as tmpf:
                write_fasta(seqs, file=tmpf)
                tmpf.flush()

            logger.info("Running all-to-all BLAST")
            run_makeblastdb(path_seqs)
            results = {(m.database, m.query): m.identity for m in run_blast(path_seqs, path_seqs)}

            logger.info("Performing clustering and dendrogram...")
            triples = sorted(
                (k1, k2, 100.0 * (1.0 - results.get((min(k1, k2), max(k1, k2)), 0.0)))
                for k1 in keys
                for k2 in keys
            )
            difference = [t[-1] for t in triples]
            dist_sq = np.asarray(difference, dtype=np.float64).reshape(len(keys), len(keys))
            dist_cd = distance.squareform(dist_sq)
            clustering = hierarchy.average(dist_cd)
            result[region] = {
                "labels": keys,
                "dist": dist_cd.tolist(),
                "linkage": clustering.tolist(),
            }

            if path_out:
                plot_phylo(clustering, keys, region, path_out % region)

    return result


def plot_phylo(linkage, labels, region, fname, *, format=None, dpi=None):
    plt.figure()
    hierarchy.dendrogram(
        linkage, labels=labels, orientation="left", link_color_func=lambda x: "black"
    )
    plt.suptitle("UPGMA for %s region" % region)
    plt.xlabel("difference [%]")
    ax = plt.gca()
    for k in ("left", "right", "top"):
        ax.spines[k].set_color("none")
    plt.tight_layout()
    plt.savefig(fname, format=format, dpi=dpi)



"""Implementation of the workflow for (single and batches of) reads.

Each read is aligned to both reference sequences.  The best match is taken and haplotyping is
performed by considering variants at the informative positions if any match could be found.
The result is a pair of ``blast.BlastMatch`` and ``haplotyping.HaplotypingResult`` (or ``None``)
that can be joined by the read file name.  Further, the sample information can be derived
from this with a regexp.
"""

import os
import re
import typing

import attr
from logzero import logger
import pandas as pd
import tempfile

from .blast import run_blast, BlastMatch
from .common import load_fasta
from .haplotyping import run_haplotyping, HaplotypingResultWithMatches

#: Default minimal quality to consider a match as true.
DEFAULT_MIN_IDENTITY = 0.5

#: Default regular expression to use for inferring the sample information.
# TODO: change a bit...?
DEFAULT_PARSE_RE = r"^(?P<sample>[^_]+_[^_]+_[^_]+)_(?P<primer>.*?)\.fasta"

#: The reference files to use.
REF_FILE = os.path.join(os.path.dirname(__file__), "data", "ref_seqs.fasta")


@attr.s(auto_attribs=True, frozen=True)
class NamedSequence:
    """A named sequence."""

    #: the sequence name
    name: str
    #: the sequence
    sequence: str


def only_blast(path_query: str) -> typing.Tuple[BlastMatch]:
    """Run BLAST and haplotyping for the one file at ``path_query``."""
    logger.info("Running BLAST on all references for %s...", path_query)
    return run_blast(REF_FILE, path_query)


def blast_and_haplotype(path_query: str) -> typing.Dict[str, HaplotypingResultWithMatches]:
    return run_haplotyping(only_blast(path_query))


def blast_and_haplotype_many(
    paths_query: typing.Iterable[str]
) -> typing.Dict[str, HaplotypingResultWithMatches]:
    """Run BLAST and haplotyping for all files at ``paths_query``.

    Return list of dicts with keys "best_match" and "haplo_result".
    """
    logger.info("Running BLAST and haplotyping for all queries...")
    result = {}
    for path_query in paths_query:
        path_result = blast_and_haplotype(path_query)
        if path_result:
            result.update(path_result)
        else:
            result[path_query] = HaplotypingResultWithMatches.build_empty()
    return result


def strip_ext(s: str) -> str:
    return s.rsplit(".", 1)[0]


def results_to_data_frames(
    results: typing.Dict[str, HaplotypingResultWithMatches], regex: str, column: str = "query"
) -> typing.Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Convert list of dicts with best_match/haplo_result to triple of Pandas DataFrame.

    The three DataFrame will contain the following information:

    1. A summary data frame showing best BLAST match target and identity plus haplotype.
    2. A data frame showing BLAST result details.
    3. A data frame showing haplotyping result details.
    """
    r_summary = []
    r_blast = []
    r_haplo = []
    for path, result in results.items():
        haplo_result = result.result
        if not haplo_result:
            for query, query_seq in load_fasta(path).items():
                r_summary.append(
                    {"query": query, "database": ".", "identity": 0, "orig_sequence": query_seq}
                )
                r_blast.append({"query": query})
                r_haplo.append({"query": query})
        else:
            query_seq = load_fasta(path)[haplo_result.query]
            haplo_matches = result.matches
            best_match = list(sorted(haplo_matches, key=lambda m: m.identity, reverse=True))[0]

            r_summary.append(
                {
                    "query": best_match.query,
                    "database": best_match.database,
                    "identity": 100.0 * best_match.identity,
                    **{
                        key: value
                        for key, value in haplo_result.asdict().items()
                        if key in ("best_haplotypes", "best_score")
                    },
                    "orig_sequence": query_seq,
                }
            )
            r_blast.append(
                {
                    "query": best_match.query,
                    "database": best_match.database,
                    "identity": 100.0 * best_match.identity,
                    "q_start": best_match.query_start,
                    "q_end": best_match.query_end,
                    "q_str": best_match.query_strand,
                    "db_start": best_match.database_start,
                    "db_end": best_match.database_end,
                    "db_str": best_match.database_strand,
                    "alignment": best_match.alignment.wrapped(
                        best_match.query_start, best_match.database_start
                    ),
                    "orig_sequence": query_seq,
                }
            )
            r_haplo.append(
                {
                    "query": best_match.query,
                    **{
                        key: value
                        for key, value in haplo_result.asdict().items()
                        if "_pos" in key
                        or "_neg" in key
                        or key in ("best_haplotypes", "best_score")
                    },
                }
            )

    dfs = pd.DataFrame(r_summary), pd.DataFrame(r_blast), pd.DataFrame(r_haplo)
    dfs = list(map(lambda df: match_sample_in_data_frame(df, regex, column), dfs))
    dfs[0] = augment_summary(
        dfs[0], results, regex, column, "sample" if "sample" in dfs[0].columns else "query"
    )
    for df in dfs:
        df.index = range(df.shape[0])
        df.insert(0, "id", df.index)
    return tuple(dfs)


SUMMARY_SUFFIX = "_ZZZ *** SUMMARY ***"


def augment_summary(
    df: pd.DataFrame,
    results: typing.Dict[str, HaplotypingResultWithMatches],
    regex: str,
    column: str,
    group_by: str,
):
    grouped = {}
    for record in results.values():
        if not record.result:
            continue
        m = re.match(regex, record.result.query)
        if m and m.group(group_by):
            if m.group(group_by) not in grouped:
                grouped[m.group(group_by)] = record.result
            else:
                grouped[m.group(group_by)] = grouped[m.group(group_by)].merge(record.result)
    rows = []
    for key, value in grouped.items():
        rows.append(
            {
                "query": "%s%s" % (key, SUMMARY_SUFFIX),
                group_by: key,
                **value.asdict(only_summary=True),
            }
        )
    for key in set(df[group_by].values) - grouped.keys():  # fill for those without matches
        rows.append({"query": "%s%s" % (key, SUMMARY_SUFFIX), group_by: key})
    orig_columns = list(df.columns.values)
    if "samples" in df.columns:
        keys = ["sample", "query"]
    else:
        keys = ["query"]
    df = df.append(pd.DataFrame(rows))[orig_columns].sort_values(keys).fillna("-")
    df["query"] = df["query"].str.replace(re.escape(SUMMARY_SUFFIX), "")
    return df


def match_sample_in_data_frame(df, regex, column):
    """Use ``regex`` to parse out information from the ``"query"`` column of the ``df`` DataFrame.

    Return an augmented DataFrame.
    """
    if not df.shape[0]:
        return df  # short-circuit empty
    # Get shortcut to column with query/filenames.
    col_query = df.loc[:, column]

    # Obtain list of new column names, sorted by occurence, must use dict for that.
    names = {}
    for query in col_query:
        m = re.match(regex, query)
        if m:
            for key in m.groupdict():
                names[key] = True
    names = list(names.keys())

    # Build new column values.
    columns = {n: [] for n in names}
    for query in col_query:
        m = re.match(regex, query)
        if m:
            for name in names:
                columns[name].append(m.groupdict().get(name))
        else:
            for name in names:
                columns[name].append(None)

    # Insert new columns into df.
    idx = df.columns.get_loc(column)
    for i, (key, column) in enumerate(columns.items()):
        df.insert(idx + i + 1, key, column)

    return df



"""Setup of Dash UI."""

import base64
import io
import os

import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_html_components as html
import dash_table
from logzero import logger
import numpy as np
import plotly.figure_factory as ff

from . import settings
from .settings import (
    MIN_IDENTITY_GREEN,
    MIN_IDENTITY_YELLOW,
    FONT_COLOR_GREEN,
    FONT_COLOR_YELLOW,
    FONT_COLOR_RED,
    BG_COLOR_GREEN,
    BG_COLOR_YELLOW,
    BG_COLOR_RED,
)
from ..haplotyping import HAPLOTYPE_NAMES
from ..phylo import plot_phylo
from .. import __version__

#: names of columns that are not to be shown
HIDDEN_COLUMNS = ("alignment", "orig_sequence")


def render_navbar():
    """Render the site navbar"""
    return dbc.Navbar(
        dbc.Container(
            children=[
                # Use row and col to control vertical alignment of logo / brand
                dbc.NavbarBrand(
                    children=[html.I(className="fas fa-carrot mr-1"), settings.HOME_BRAND],
                    id="page-brand",
                ),
                dbc.Nav(
                    dbc.NavItem(
                        dbc.NavLink(
                            dcc.Upload(
                                id="upload-data",
                                children=[
                                    html.I(className="fas fa-cloud-upload-alt mr-1"),
                                    "upload files",
                                ],
                                multiple=True,
                            ),
                            className="btn btn-outline-secondary",
                            active=True,
                        )
                    )
                ),
            ]
        ),
        color="primary",
        dark=True,
        id="page-navbar",
    )


def render_page_content_empty_children():
    """Return page content when there's no content (no sequence has been uploaded yet)."""
    return [
        html.Div(
            children=[
                html.P(
                    children=[
                        "Click ",
                        html.Span(
                            children=[
                                html.I(className="fas fa-cloud-upload-alt mr-1"),
                                "upload files",
                            ],
                            className="badge badge-secondary ml-1 mr-1",
                        ),
                        'upload files on the top right to upload one or more .fasta, .fastq, "'
                        ".scf, or .ab1 files.  You can find a ZIP file with some examples ",
                        html.A(
                            children=[html.I(className="fas fa-download mr-1"), "here"],
                            href="/dash/assets/hlso_example.zip",
                            className="badge badge-primary",
                        ),
                        ".",
                    ]
                ),
                html.P(
                    children=[
                        "You can find the Haplotype-Lso ",
                        html.A(
                            children=[html.I(className="fa fa-book"), " manual here"],
                            href="https://haplotype-lso.readthedocs.io/en/latest",
                        ),
                        " which also contains the ",
                        html.A(
                            children=[html.I(className="fa fa-shoe-prints"), " Tutorial"],
                            href="https://haplotype-lso.readthedocs.io/en/latest/overview_tutorial.html",
                        ),
                        ".",
                    ]
                ),
                html.P(
                    children=[html.I(className="fa fa-smile-wink"), " Happy haplotyping!"],
                    className="text-center",
                ),
            ]
        ),
        html.Div(children=[dash_table.DataTable(id="blast-table")], style={"display": "none"}),
        html.Div(
            children=[dash_table.DataTable(id="haplotyping-table")], style={"display": "none"}
        ),
        html.Div(children=["placeholder"], id="blast-current-match", style={"display": "none"}),
    ]


def render_page_content_empty():
    return html.Div(
        children=[
            html.Div(render_page_content_empty_children(), className="text-center text-muted")
        ],
        id="page-content",
    )


def render_main_content():
    """Render page main content"""
    return html.Div(
        children=[dbc.Row(dbc.Col(children=[render_page_content_empty()]))],
        className="container pt-3",
    )


def render_footer():
    """Render page footer"""
    return html.Footer(
        html.Div(
            children=[
                html.Div(
                    children=[
                        html.Div(
                            children=[
                                html.Span("haplotype-lso v%s" % __version__, className="text-muted")
                            ],
                            className="col-6",
                        ),
                        html.Div(
                            children=[
                                html.A(
                                    children=[
                                        html.I(className="fab fa-github mr-1"),
                                        "GitHub Project",
                                    ],
                                    href="https://github.com/holtgrewe/haplotype-lso",
                                    className="text-muted",
                                )
                            ],
                            className="col-6 text-right",
                        ),
                    ],
                    className="row",
                )
            ],
            className="container",
        ),
        className="footer",
    )


def render_tab_summary(session_data):
    logger.info("Rendering Summary Tab")
    col_ids = list(session_data["blast"].columns)
    columns = [{"name": col, "id": col} for col in col_ids]
    style_cell = {"font-family": "sans-serif"}
    style_cell_conditional = [
        {"if": {"column_id": "query"}, "textAlign": "left"},
        {"if": {"column_id": "subject"}, "textAlign": "left"},
    ]
    style_data_conditional = [
        {
            "if": {
                "column_id": "identity",
                "filter_query": "{identity} < %f" % MIN_IDENTITY_YELLOW,
            },
            "color": FONT_COLOR_RED,
            "backgroundColor": BG_COLOR_RED,
        },
        {
            "if": {
                "column_id": "identity",
                "filter_query": "{identity} >= %f" % MIN_IDENTITY_YELLOW,
            },
            "color": FONT_COLOR_YELLOW,
            "backgroundColor": BG_COLOR_YELLOW,
        },
        {
            "if": {
                "column_id": "identity",
                "filter_query": "{identity} >= %f" % MIN_IDENTITY_GREEN,
            },
            "color": FONT_COLOR_GREEN,
            "backgroundColor": BG_COLOR_GREEN,
        },
    ]
    style_header = {"text-align": "center", "fontWeight": "bold"}
    df = session_data["summary"]
    df.loc[:, "identity"] = df.loc[:, "identity"].map(
        lambda x: str(round(float(x), 1)) if x != "-" else x
    )
    table = dash_table.DataTable(
        id="summary-table",
        columns=[{"name": i, "id": i} for i in df.columns if i not in HIDDEN_COLUMNS],
        data=[
            {k: v for k, v in record.items() if k not in HIDDEN_COLUMNS}
            for record in df.to_dict("records")
        ],
        style_cell=style_cell,
        style_cell_conditional=style_cell_conditional,
        style_data_conditional=style_data_conditional,
        style_header=style_header,
    )
    return [dcc.Loading(table)]


def render_tab_blast(session_data):
    logger.info("Rendering BLAST Tab")
    col_ids = list(session_data["blast"].columns)
    columns = [{"name": col, "id": col} for col in col_ids]
    style_cell = {"font-family": "sans-serif"}
    style_cell_conditional = [
        {"if": {"column_id": "query"}, "textAlign": "left"},
        {"if": {"column_id": "subject"}, "textAlign": "left"},
    ]
    style_data_conditional = [
        {
            "if": {
                "column_id": "identity",
                "filter_query": "{identity} < %f" % MIN_IDENTITY_YELLOW,
            },
            "color": FONT_COLOR_RED,
            "backgroundColor": BG_COLOR_RED,
        },
        {
            "if": {
                "column_id": "identity",
                "filter_query": "{identity} >= %f" % MIN_IDENTITY_YELLOW,
            },
            "color": FONT_COLOR_YELLOW,
            "backgroundColor": BG_COLOR_YELLOW,
        },
        {
            "if": {
                "column_id": "identity",
                "filter_query": "{identity} >= %f" % MIN_IDENTITY_GREEN,
            },
            "color": FONT_COLOR_GREEN,
            "backgroundColor": BG_COLOR_GREEN,
        },
    ]
    style_header = {"text-align": "center", "fontWeight": "bold"}
    df = session_data["blast"]
    df.loc[:, "identity"] = df.loc[:, "identity"].map(lambda x: str(round(float(x), 1)))
    table = dash_table.DataTable(
        id="blast-table",
        columns=[{"name": i, "id": i} for i in df.columns if i not in HIDDEN_COLUMNS],
        data=[
            {k: v for k, v in record.items() if k not in HIDDEN_COLUMNS}
            for record in df.to_dict("records")
        ],
        style_cell=style_cell,
        style_cell_conditional=style_cell_conditional,
        style_data_conditional=style_data_conditional,
        style_header=style_header,
        row_selectable="single",
    )
    blast_match_div = html.Div(id="blast-current-match")
    return [dcc.Loading(table), dcc.Loading(blast_match_div)]


def _pos_neg(x):
    return x.replace("_neg", "-").replace("_pos", "+")


def render_tab_haplotyping(session_data):
    logger.info("Rendering Haplotying Tab")
    df = session_data["haplotyping"]
    columns = [{"name": col, "id": col} for col in df.columns]
    style_cell = {"font-family": "sans-serif"}
    style_header = {"text-align": "center", "fontWeight": "bold"}
    table = dash_table.DataTable(
        id="haplotyping-table",
        columns=[{"name": _pos_neg(i), "id": _pos_neg(i)} for i in df.columns],
        data=[
            {_pos_neg(key): value for key, value in record.items()}
            for record in df.to_dict("records")
        ],
        style_cell=style_cell,
        style_header=style_header,
    )
    return [table]


def render_tab_dendrograms(session_data):
    logger.info("Rendering Dendrogram Tab")
    result = []

    for region, entry in session_data["phylo"].items():
        result.append(html.H3("%s Region" % region))
        if "message" in entry:
            result.append(html.P(entry["message"]))
        else:
            labels = entry["labels"]
            linkage = np.asarray(entry["linkage"])
            buf = io.BytesIO()
            plot_phylo(linkage, labels, region, buf, format="png", dpi=300)
            result.append(
                html.Img(
                    src="data:image/png;base64,%s"
                    % base64.b64encode(buf.getvalue()).decode("utf-8"),
                    width=640,
                    height=480,
                )
            )

    return result


def render_hidden():
    return html.Div(id="hidden-data", style={"display": "none"})


def build_layout():
    """Build the overall Dash app layout"""
    return html.Div(
        children=[
            # Represents the URL bar, doesn't render anything.
            dcc.Location(id="url", refresh=False),
            # Navbar, content, footer.
            render_navbar(),
            render_main_content(),
            render_footer(),
            render_hidden(),
        ],
        id="_dash-app-content",  # make pytest-dash happy
    )



def load_tsv(input_path):
    header = None
    records = []
    with open(input_path, "rt") as inputf:
        for line in inputf:
            if line.startswith("#"):
                continue
            arr = line.strip().split("\t")
            if not header:
                header = arr
            else:
                records.append(dict(zip(header, arr)))

    return header, records


def rev(seq):
    return list(reversed(seq))


def revcomp(seq):
    m = {"a": "t", "A": "T", "t": "a", "T": "A", "c": "g", "C": "G", "g": "c", "G": "C"}
    return "".join(reversed(list(map(lambda x: m.get(x, x), seq))))


def write_fasta(seqs, file):
    for name, seq in seqs.items():
        print(">%s\n%s" % (name, seq), file=file)


def load_fasta(path):
    result = {}
    with open(path, "rt") as inputf:
        name = None
        seq_lines = []
        for line in inputf:
            line = line.rstrip()
            if line.startswith(">"):
                if name:
                    result[name] = "".join(seq_lines)
                name = line[1:].split()[0]
                seq_lines = []
            elif name:  # ignore if first line is not ">"
                seq_lines.append(line)
        if name:
            result[name] = "".join(seq_lines)
    return result


def describe(*, pos, ref, alt, **kwargs):
    """Return description of variant."""
    if "-" in ref and "-" in alt:
        raise Exception("Cannot handle this yet!")
    elif "-" in ref:
        if ref[1] != "-":
            if len(ref) == 1:
                return "n.%ddel%sins%s" % (pos + 1, only_bases(ref[1:]), alt[1:])
            else:
                return "n.%d_%ddel%sins%s" % (
                    pos + 1,
                    pos + len(ref) - 1,
                    only_bases(ref[1:]),
                    alt[1:],
                )
        else:
            return "n.%d_%dins%s" % (pos, pos + 1, alt[1:])
    elif "-" in alt:
        if alt[1] != "-":
            if len(ref) == 1:
                return "n.%ddel%sins%s" % (pos + 1, ref[1:], only_bases(alt[1:]))
            else:
                return "n.%d_%ddel%sins%s" % (
                    pos + 1,
                    pos + len(ref) - 1,
                    ref[1:],
                    only_bases(alt[1:]),
                )
        else:
            if len(ref) == 2:
                return "n.%ddel%s" % (pos + 1, ref[1:])
            else:
                return "n.%d_%ddel%s" % (pos + 1, pos + len(ref) - 1, ref[1:])
    elif len(ref) == 1 and len(alt) == 1:
        return "n.%d%s>%s" % (pos, ref, alt)
    else:
        if len(ref) == 2:
            return "n.%ddel%s" % (pos + 1, ref[1:])
        else:
            return "n.%d_%ddel%s" % (pos + 1, pos + len(ref) - 1, ref[1:])


def only_bases(s):
    return "".join([x for x in s if x.upper() in "CGATN"])


def normalize_var(ref_seq, curr_var, offset):
    """Normalize the variant."""
    ref_seq = only_bases(ref_seq)
    ref = only_bases(curr_var["ref_bases"])
    alt = only_bases(curr_var["alt_bases"])
    orig_ref, orig_alt = ref, alt
    pos = curr_var["pos"] - 1 - offset
    delta = 0
    while ref[-1] == alt[-1] and pos > 0:
        delta += 1
        pos -= 1
        ref = ref_seq[pos] + ref
        alt = ref_seq[pos] + alt
        while ref[-1] == alt[-1] and len(ref) > 1 and len(alt) > 1:
            ref = ref[:-1]
            alt = alt[:-1]

    result = dict(curr_var)
    result["pos"] -= delta
    result["ref"] = ref if len(ref) >= len(alt) else ref + "-" * (len(alt) - len(ref))
    result["alt"] = alt if len(ref) <= len(alt) else alt + "-" * (len(ref) - len(alt))
    result["ref_bases"] = ref
    result["alt_bases"] = alt
    result["description"] = describe(**result)
    return result


def call_variants(ref, alt, offset=0):
    """Call variants from alignment (``ref`` and ``alt`` with gaps).

    The resulting positions are 1-based.
    """
    if len(ref) != len(alt):
        raise Exception("Invalid alignment row lengths: %d vs. %d", len(ref), len(alt))
    result = {}
    ref = ref.upper()
    alt = alt.upper()

    pos_ref = offset  # position in ref
    i = 0  # position in rows
    curr_var = {}
    while i < len(ref):
        if ref[i] == alt[i]:
            if curr_var:
                if curr_var["ref"].startswith("-") or curr_var["alt"].startswith("-"):
                    if curr_var["ali_pos"] > 1:  # left-pad indels if possible
                        curr_var = {
                            "pos": curr_var["pos"] - 1,
                            "ali_pos": curr_var["ali_pos"] - 1,
                            "ref": ref[curr_var["ali_pos"] - 2] + curr_var["ref"],
                            "alt": alt[curr_var["ali_pos"] - 2] + curr_var["alt"],
                        }
                    else:  # use base right of it, VCF-style
                        curr_var = {
                            "pos": curr_var["pos"],
                            "ali_pos": curr_var["ali_pos"],
                            "ref": curr_var["ref"]
                            + ref[curr_var["ali_pos"] + len(curr_var["ref"])],
                            "alt": curr_var["alt"]
                            + alt[curr_var["ali_pos"] + len(curr_var["alt"])],
                        }
                curr_var["ref_bases"] = only_bases(curr_var["ref"])
                curr_var["alt_bases"] = only_bases(curr_var["alt"])
                curr_var["description"] = describe(**curr_var)
                curr_var = normalize_var(ref, curr_var, offset)
                result[curr_var["pos"]] = curr_var
                curr_var = {}
        else:
            if curr_var:
                curr_var["ref"] = curr_var["ref"] + ref[i]
                curr_var["alt"] = curr_var["alt"] + alt[i]
            else:
                curr_var = {"pos": pos_ref + 1, "ali_pos": i + 1, "ref": ref[i], "alt": alt[i]}

        if ref[i] != "-":
            pos_ref += 1
        i += 1

    if curr_var:
        if curr_var["ref"].startswith("-") or curr_var["alt"].startswith("-"):
            if curr_var["ali_pos"] > 1:  # left-pad indels if possible
                curr_var = {
                    "pos": curr_var["pos"] - 1,
                    "ali_pos": curr_var["ali_pos"] - 1,
                    "ref": ref[curr_var["ali_pos"] - 2] + curr_var["ref"],
                    "alt": alt[curr_var["ali_pos"] - 2] + curr_var["alt"],
                }
            else:  # use base right of it, VCF-style
                curr_var = {
                    "pos": curr_var["ali_pos"],
                    "ali_pos": curr_var["ali_pos"],
                    "ref": curr_var["ref"] + ref[curr_var["ali_pos"] + len(curr_var["ref"])],
                    "alt": curr_var["alt"] + alt[curr_var["ali_pos"] + len(curr_var["alt"])],
                }
        curr_var["ref_bases"] = only_bases(curr_var["ref"])
        curr_var["alt_bases"] = only_bases(curr_var["alt"])
        curr_var = normalize_var(ref, curr_var, offset)
        curr_var["description"] = describe(**curr_var)
        result[curr_var["pos"]] = curr_var

    return result



"""Helpers for converting read files."""

import os
import tempfile
import typing

from bioconvert.scf2fasta import SCF2FASTA
from bioconvert.abi2fasta import ABI2FASTA
from bioconvert.fastq2fasta import FASTQ2FASTA
from logzero import logger

from .common import load_fasta


def convert_seqs(
    seq_files: typing.Iterable[str], tmpdir: str, sample_name_from_file_name: bool = False
) -> typing.List[str]:
    """Convert SRF and AB1 files to FASTQ."""
    logger.info("Running file conversion...")
    result = []

    for seq_file in seq_files:
        file_basename = os.path.basename(seq_file)[: -len(".fasta")]
        path_fasta = os.path.join(tmpdir, file_basename) + ".fasta"
        result.append(path_fasta)

        path_fasta_tmp = tempfile.NamedTemporaryFile(dir=tmpdir, delete=False)
        if seq_file.endswith(".scf"):
            logger.info("Converting SCF file %s...", seq_file)
            SCF2FASTA(seq_file, path_fasta_tmp)()
            result.append(path_fasta_tmp)
        elif seq_file.endswith(".ab1"):
            logger.info("Converting ABI file %s...", seq_file)
            ABI2FASTA(seq_file, path_fasta_tmp)()
            result.append(path_fasta_tmp)
        elif seq_file.endswith(".fastq"):
            logger.info("Converting FASTAQ file %s...", seq_file)
            FASTQ2FASTA(seq_file, path_fasta_tmp)()
            result.append(path_fasta_tmp)
        else:
            path_fasta_tmp = seq_file

        fasta_content = load_fasta(path_fasta_tmp)

        if sample_name_from_file_name and len(fasta_content) != 1:
            prefix_no = 1
        else:
            prefix_no = 0

        with open(path_fasta, "wt") as outputf:
            for name, seq in fasta_content.items():
                if sample_name_from_file_name:
                    name = file_basename
                if prefix_no:
                    prefix = "_%d" % prefix_no
                    prefix_no += 1
                else:
                    prefix = ""
                print(">%s%s\n%s" % (prefix, name, seq), file=outputf)

    logger.info("Done converting files.")
    return result



"""Download of seed references."""

import os
import textwrap

from Bio import Entrez
from logzero import logger

from .cli import _proc_args
from .common import load_tsv

#: Known FASTA extensions.
FASTA_EXTS = (".fasta", ".fa", ".fna")

#: Reference sequences.
REF_SEQS = {
    "16S": {"accession": "EU812559.1", "start": 0, "end": 1225},
    "16S-23S": {"accession": "EU812559.1", "start": 1225, "end": 2515},
    "50S": {"accession": "EU834131.1", "start": 0, "end": 1714},
}


def download_references(_parser, args):
    """Download references."""
    out_dir = os.path.dirname(args.out_tsv)

    logger.info("Downloading references...")
    for name, vals in REF_SEQS.items():
        out_path_fasta = os.path.join(out_dir, "ref_" + name + ".fasta")
        if os.path.exists(out_path_fasta):
            logger.warn("- already exists: %s", vals["accession"])
        else:
            logger.info("- downloading: %s", vals["accession"])
            res = Entrez.efetch(
                db="nucleotide", id=vals["accession"], rettype="fasta", retmode="text"
            )
            header, seq = res.read().rstrip().split("\n", 1)
            seq = "".join([c for c in seq if c in "cgatnCGATN"])
            prefix = "N" * vals["start"]
            suffix = "N" * (len(seq) - vals["end"])
            seq = prefix + seq[vals["start"] : vals["end"]] + suffix
            with open(out_path_fasta, "wt") as outputf:
                print(header, file=outputf)
                print("\n".join(textwrap.wrap(seq)), file=outputf)


def download_seeds(parser, args):
    """Download seed sequences"""
    out_dir = os.path.dirname(args.out_tsv)
    out_path_tsv = os.path.join(out_dir, "seeds_paths.tsv")
    if os.path.exists(out_path_tsv):
        logger.warn("Output file already exists: %s", out_path_tsv)
        logger.warn("Skipping reference download.")
        return

    logger.info("Loading input TSV...")
    header, records = load_tsv(args.in_tsv)

    logger.info("Downloading sequences...")
    for record in records:
        if os.path.exists(os.path.join(out_dir, record["accession"])):
            logger.info("- already exists: %s", os.path.join(out_dir, record["accession"]))
            record["path"] = record["accession"]
            out_path = os.path.join(out_dir, record["accession"])
            if not os.path.exists(out_path):
                parser.exit(1, "File not found: %s" % out_path)
        else:
            logger.info("- downloading: %s", record["accession"])
            record["path"] = record["accession"] + ".fasta"
            out_path_fasta = os.path.join(out_dir, record["path"])
            res = Entrez.efetch(
                db="nucleotide", id=record["accession"], rettype="fasta", retmode="text"
            )
            record = res.read()
            with open(out_path_fasta, "wt") as outputf:
                outputf.write(record.rstrip() + "\n")

    logger.info("Writing output TSV: %s", out_path_tsv)
    header[-1] = "path"
    with open(out_path_tsv, "wt") as outputf:
        print("\t".join(header), file=outputf)
        for record in records:
            print("\t".join([record[k] for k in header]), file=outputf)


def run(parser, args):
    """Perform the download of seed sequences and references."""
    logger.info("Starting download of seed sequences and reference.")
    logger.info("Arguments are %s", vars(args))

    download_references(parser, args)
    download_seeds(parser, args)

    logger.info("All done. Have a nice day!")


def add_parser(subparser):
    """Configure the ``argparse`` sub parser."""
    parser = subparser.add_parser("ref_download")
    parser.set_defaults(func=run)

    parser.add_argument("in_tsv", help="Path to output TSV file.")
    parser.add_argument("out_tsv", help="Path to output TSV file.")



"""Code for the web interface to Haplotype-Lso."""

import os

from logzero import logger

from . import settings


def run(parser, args):
    """Run the ``hlso`` command line interface."""
    logger.info("Configuring settings from arguments %s", args)
    settings.HOST = args.host
    settings.PORT = args.port
    settings.PUBLIC_URL_PREFIX = args.public_url_prefix

    logger.info("Running server...")
    from .app import app  # noqa

    app.run_server(
        host=settings.HOST, port=settings.PORT, debug=args.debug, dev_tools_hot_reload=args.debug
    )
    logger.info("Web server stopped. Have a nice day!")


def add_parser(subparser):
    """Configure the ``argparse`` sub parser."""
    parser = subparser.add_parser("web")
    parser.set_defaults(func=run)

    parser.add_argument(
        "--host", help="Server host", default=os.environ.get("HLSO_HOST", "0.0.0.0")
    )
    parser.add_argument(
        "--port", help="Server port", type=int, default=int(os.environ.get("HLSO_PORT", "8050"))
    )
    parser.add_argument(
        "--public-url-prefix",
        default=os.environ.get("HLSO_URL_PREFIX", ""),
        help="The prefix that this app will be served under (e.g., if behind a reverse proxy.)",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=(os.environ.get("DEBUG", "0").lower() in ("1", "true", "yes")),
        help="Whether or not to enable debugging.",
    )



"""Code for running blastn and handling matches.

We store matches in ``BlastMatch`` objects that can be easily converted into dicts and
subsequently Pandas dataframe rows.
"""

import os
import subprocess
import typing
import shlex
import json

import attr
from Bio.Blast import NCBIXML
import io
from logzero import logger

from .common import revcomp, rev


@attr.s(auto_attribs=True, frozen=True)
class Alignment:
    """Representation of an alignment."""

    #: (database) hit sequence
    hseq: str
    #: alignment mid line
    midline: str
    #: query sequence
    qseq: str

    def revcomp(self) -> typing.TypeVar("Alignment"):
        return Alignment(revcomp(self.hseq), "".join(reversed(self.midline)), revcomp(self.qseq))

    def wrapped(self, qry_start: int, db_start: int, line_length: int = 60) -> str:
        result = []
        for offset in range(0, len(self.hseq), line_length):
            end = min(len(self.hseq), offset + line_length)
            result += [
                ("Sbjct %%4d %%-%ds %%s" % line_length)
                % (db_start + offset + 1, self.hseq[offset:end], db_start + offset + line_length),
                ("      %%4s %%-%ds" % line_length) % ("", "".join(self.midline[offset:end])),
                ("Query %%4d %%-%ds %%s" % line_length)
                % (qry_start + offset + 1, self.qseq[offset:end], qry_start + offset + line_length),
                "",
            ]
        return "\n".join(result)

    @staticmethod
    def build_empty() -> typing.TypeVar("Alignment"):
        return Alignment(hseq="", midline="", qseq="")


@attr.s(auto_attribs=True, frozen=True)
class BlastMatch:
    """Representation of a match."""

    #: query file name
    path: typing.Optional[str]
    #: query sequence name
    query: str
    #: database sequence name
    database: str
    #: identity fraction
    identity: float
    #: bit score
    bits: float
    #: strand of query ("+" or "-")
    query_strand: str
    #: 0-based start position
    query_start: int
    #: 0-based end position
    query_end: int
    #: strand of database ("+" or "-")
    database_strand: str
    #: 0-based start position
    database_start: int
    #: 0-based end position
    database_end: int
    #: CIGAR string of match
    match_cigar: str
    #: matching sequence
    match_seq: str
    #: Alignment
    alignment: typing.Optional[Alignment]

    @property
    def database_length(self):
        return self.database_end - self.database_start

    @property
    def query_length(self):
        return self.query_end - self.query_start

    @property
    def is_match(self) -> bool:
        return bool(self.database)

    @staticmethod
    def build_nomatch(query, path: str = None) -> typing.TypeVar("BlastMatch"):
        return BlastMatch(
            path=path,
            query=query,
            database=None,
            identity=0.0,
            bits=0.0,
            query_strand=".",
            query_start=0,
            query_end=0,
            database_strand=".",
            database_start=0,
            database_end=0,
            match_cigar="",
            match_seq="",
            alignment=Alignment.build_empty(),
        )


def is_nucl(c: str) -> bool:
    """Helper, returns wether ``c`` is a nucleotide character."""
    return c in "acgtnACGTN"


def match_cigar(
    qseq: str, hseq: str, query_start: int, query_end: int, query_len: int
) -> typing.Tuple[typing.Tuple[int, str]]:
    """Return CIGAR string for the match with the given paraemters."""
    match_cigar = []
    if query_start > 0:
        match_cigar.append([query_start, "H"])
    for q, h in zip(qseq, hseq):
        if q == "-":
            op = "D"
        elif h == "-":
            op = "I"
        else:
            op = "M"
        if match_cigar and match_cigar[-1][1] == op:
            match_cigar[-1][0] += 1
        else:
            match_cigar.append([1, op])
    if query_end != query_len:
        match_cigar.append([query_len - query_end, "H"])
    return tuple(tuple(t) for t in match_cigar)


def only_dna(seq: str) -> str:
    """Return ``str`` with all ACGTN characters from ``seq``."""
    return "".join(filter(lambda x: x in "ACGTNacgtn", seq))


def parse_blastn_xml(
    blastn_xml: str, path_query: typing.Optional[str] = None
) -> typing.Tuple[BlastMatch]:
    """Parse BLASTN output in JSON format and return list of ``BlastMatch``."""
    result = []
    for blast_record in NCBIXML.parse(io.StringIO(blastn_xml)):
        for blast_alignment in blast_record.alignments:
            hsp = blast_alignment.hsps[0]
            qseq = only_dna(hsp.query)
            hseq = only_dna(hsp.sbjct)
            db_strand = "+" if hsp.strand[1] == "Plus" else "-"
            db_start = min(hsp.sbjct_start, hsp.sbjct_end) - 1
            db_end = max(hsp.sbjct_start, hsp.sbjct_end)
            query_strand = "+" if hsp.strand[0] == "Plus" else "-"
            query_start = min(hsp.query_start, hsp.query_end) - 1
            query_end = max(hsp.query_start, hsp.query_end)
            if db_strand == "-":
                db_strand = "+"
                query_strand = "-" if query_strand == "+" else "+"
                qseq = revcomp(qseq)
                hseq = revcomp(hseq)
                alignment = Alignment(
                    hseq=revcomp(hsp.sbjct), midline=rev(hsp.match), qseq=revcomp(hsp.query)
                )
            else:
                alignment = Alignment(hseq=hsp.sbjct, midline=hsp.match, qseq=hsp.query)
            cigar = match_cigar(qseq, hseq, query_start, query_end, blast_record.query_letters)
            result.append(
                BlastMatch(
                    path=path_query,
                    query=blast_record.query,
                    database=blast_alignment.title.split()[1],
                    bits=hsp.bits,
                    identity=hsp.identities / hsp.align_length,
                    query_strand=query_strand,
                    query_start=query_start,
                    query_end=query_end,
                    database_strand=db_strand,
                    database_start=db_start,
                    database_end=db_end,
                    match_cigar="".join(["".join(map(str, x)) for x in cigar]),
                    match_seq="".join([x for x in filter(is_nucl, qseq)]),
                    alignment=alignment,
                )
            )
    return tuple(result)


def run_blast(database: str, query: str) -> typing.Tuple[BlastMatch]:
    """Run blastn on FASTA query ``query`` to database sequence at ``database``."""
    cmd = ("blastn", "-db", database, "-query", query, "-outfmt", "16")
    logger.info("Executing %s", repr(" ".join(cmd)))
    return parse_blastn_xml(subprocess.check_output(cmd).decode("utf-8"), path_query=query)


def run_makeblastdb(path: str, dbtype: str = "nucl") -> str:
    """Run blastn on FASTA query ``query`` to database sequence at ``database``."""
    cmd = ("makeblastdb", "-in", path, "-dbtype", "nucl")
    logger.info("Executing %s", repr(" ".join(cmd)))
    return subprocess.check_output(cmd).decode("utf-8")



"""Haplotype-Lso web front-end settings."""

# Not Configurable ===========================================================

#: String to display on home screen.
HOME_BRAND = "Haplotype-Lso"

# TODO: should this come from the UI?
#: The sample regex
SAMPLE_REGEX = r"^(?P<sample>[^\.]+)\.(?P<region>[^\.]+)(?:\.(?P<primer>.*))?"

#: Whether to write file name as sample name.
FILE_NAME_TO_SAMPLE_NAME = True

#: Known regions.
KNOWN_REGIONS = ("16S", "16S-23S", "50S")
#: Known primers.
KNOWN_PRIMERS = ("LsoF", "Ol2cR", "LsoTX1623F", "LsoTX1623R", "Cl514F", "Cl514R")

#: Font color for red conditional formatting.
FONT_COLOR_RED = "#CC0000"
#: Background color for red conditional formatting.
BG_COLOR_RED = "#FFCCCC"
#: Font color for yellow conditional formatting.
FONT_COLOR_YELLOW = "#996600"
#: Background color for yellow conditional formatting.
BG_COLOR_YELLOW = "#FFFFCC"
#: Font color for green conditional formatting.
FONT_COLOR_GREEN = "#006600"
#: Background color for green conditional formatting.
BG_COLOR_GREEN = "#CCFFCC"

#: Minimal identity (in percent) for green color.
MIN_IDENTITY_GREEN = 98
#: Minimal identity (in percent) for yellow color.
MIN_IDENTITY_YELLOW = 96


# Configurable ===============================================================

#: The host to listen on.
HOST = "0.0.0.0"
#: The port to listen on.
PORT = 8050

#: The public URL prefix to use.
PUBLIC_URL_PREFIX = ""
